url = "http://localhost:6363/"
json_context = {
    "_": "_:",
    "doc": "terminusdb:///system/data/",
    "layer": "http://terminusdb.com/schema/layer#",
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "ref": "http://terminusdb.com/schema/ref#",
    "repo": "http://terminusdb.com/schema/repository#",
    "terminus": "http://terminusdb.com/schema/terminus#",
    "vio": "http://terminusdb.com/schema/vio#",
    "woql": "http://terminusdb.com/schema/woql#",
    "xdd": "http://terminusdb.com/schema/xdd#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
}

dbrec = {
    "db": "aaaaaa",
    "account": "admin",
    "title": "admin|aaaaaa",
    "description": "dasdasdds",
}

dbrecs = [
    {"db": "5534534", "account": "admin", "title": "3453", "description": "345345"},
    {
        "db": "adsasasddsa",
        "account": "admin",
        "title": "admin|adsasasddsa",
        "description": "asdsadsda",
    },
    {"db": "blah", "account": "admin", "title": "admin|blah", "description": "adfadf"},
    {
        "db": "daassd",
        "account": "admin",
        "title": "admin|daassd",
        "description": "asdasdsd",
    },
    {
        "db": "dassadds",
        "account": "admin",
        "title": "admin|dassadds",
        "description": "asdsdsad",
    },
    {"db": "ddd", "account": "admin", "title": "admin|ddd", "description": "ddds"},
    {"db": "ffff", "account": "admin", "title": "admin|ffff", "description": "fffff"},
    {
        "db": "fffffff",
        "account": "admin",
        "title": "admin|fffffff",
        "description": "fff",
    },
    {
        "db": "ggggggggg",
        "account": "admin",
        "title": "admin|ggggggggg",
        "description": "gg",
    },
    {
        "db": "twretwert",
        "account": "admin",
        "title": "admin|twretwert",
        "description": "adfadf",
    },
    {
        "db": "terminus",
        "account": "",
        "title": "Master Database",
        "description": "The master database contains the meta-data about databases, users and roles",
    },
]
